
from abc import ABC
from typing import Dict, List, Iterable

from relationalai.debugging import logger
from relationalai.dsl import next_id

from relationalai.experimental.pathfinder.options import DEFAULT_OPTIONS
from relationalai.experimental.pathfinder.rpq import RPQ
from relationalai.experimental.pathfinder.filter import (
    NodeFilter, NodeLabel, EdgeLabel, AnonymousNodeFilter
)
from relationalai.experimental.pathfinder.automaton import prepare_automaton
from relationalai.experimental.pathfinder.transition import EdgeTransition, NodeTransition

from relationalai.experimental.pathfinder.datalog import (
    ConstantEqualityAtom, Rule, RelAtom, AnonymousFilterAtom,
    VariableEqualityAtom, Atom, DatalogProgram
)

# =========================================================================================
# Compiling a pattern into Rel programs
# =========================================================================================
#
# We compile a RPQ pattern into two kinds of programs:
# 1. A program that computes the connectivity relation of the input automaton.
# 2. A program that computes the product graph of the input automaton and a knowledge graph.
#
# We illustrate the compilation process on the following example pattern:
# -----------------------------------------------------------------------------------------
# Running example
# -----------------------------------------------------------------------------------------
# An example of the program generation is presented on the pattern below.
#
# path(Odd, star('-[A]->', Even)) | path('<-[B]-', N)
#
# which is parsed into the following RPQ
#
# ╷ Union
# ├──┐ Concat
# │  ├──╴ Node(Odd)
# │  └──┐ Star
# │     └──┐ Concat
# │        ├──╴ Edge(-[A]→) [pos=1]
# │        └──╴ Node(Even)
# └──┐ Concat
#    ├──╴ Edge(←[B]-) [pos=2]
#    └──╴ Node(N)
#
# The RPQ is compiled into the following automaton (with transitions explicitly labeled):
# * states: 0 1 2
# * init: 0
# * delta:
#    - 0
#       → 1
#         ⊨(Odd)=[-[A]→]=()⇒    (t1)
#       → 2
#         ⊨()=[←[B]-]=()⇒       (t2)
#    - 1
#       → 1
#         ⊨(Even)=[-[A]→]=()⇒   (t3)
# * final:
#    - 0
#       ⊨(Odd)⇒                 (f1)
#    - 1
#       ⊨(Even)⇒                (f2)
#    - 2
#       ⊨(N)⇒                   (f3)
#
# =========================================================================================
# 1. Connectivity Program
# =========================================================================================
# The Rel connectivity program defines a binary relation `pq_conn(x, y)` that selects the
# nodes `y` reachable from `x` with the automaton having an accepting run. The program has
# two independent uses:
# 1. When the RPQ is meant to be evaluated as a connectivity query, the program is used to
#    evaluate the query.
# 2. When the RPQ is meant to produce paths but the user does not specify the source or
#    target nodes to be used during evaluation.
#
# The program consists of binary predicates `pq_conn_from_q(x, y)` that selects the nodes
# `y` reachable from `x` with the automaton having a run from state `q` to an accepting
# state. Note that when the automaton has cycles, the program may be recursive. To make sure
# that each introduced predicate is grounded we may "contract" the final node transitions of
# an accepting state `q` with every incoming transition of `q` (note that the final node
# transitions may be ungrounded). As a consequence, an accepting state `q` that does not
# have any outgoing transitions, called an _exit state_, does not need its own dedicated
# `pq_conn_fom_q` predicate. At this time, we decorate every predicate with `@no_inline`
# attribute to prevent any inlining. Diligent inlining may be beneficial for performance but
# is left for future work.
#
# -----------------------------------------------------------------------------------------
# In the running example, the state 2 is an exit state. Also note that the initial state is
# also a final state, thus the automaton accepts empty paths (consisting of a single node
# with type OddNode). The compiled Rel program follows. (All relations have `@no_inline`
# attribute that we omit here for brevity)
#
# // The main connectivity predicate
# def pq_conn(x0, x1): pq_conn_from_0(x0, x1)
#
# // Accepting empty paths with node transition f1
# def pq_conn(x0, x1): x0 = x1 ∧ Odd(x0)
#
# // Paths from state 0 going through t1 to state 1
# def pq_conn_from_0(x0, x2): A(x0, x1) ∧ Odd(x0) ∧ pq_conn_from_1(x1, x2)
#
# // Paths from state 0 through t1 (to state 1) and terminating with f2
# def pq_conn_from_0(x0, x1): A(x0, x1) ∧ Odd(x0) ∧ Even(x1)
#
# // Paths from state 0 through t2 (to state 1) and terminating with f3
# def pq_conn_from_0(x0, x1): A(x0, x1) ∧ Even(x0) ∧ Even(x1)
#
# // Paths from state 1 through t3 to state 1
# def pq_conn_from_1(x0, x2): A(x0, x1) ∧ Even(x0) ∧ pq_conn_from_1(x1, x2)
#
# // Paths from state 1 through t3 and terminating with f2
# def pq_conn_from_1(x0, x1): A(x0, x1) ∧ Even(x0) ∧ Even(x1)
#
# =========================================================================================
# 2. Product Graph Program
# =========================================================================================
# The Product Graph program defines a relation `pg_graph` of arity 5 that defines the set of
# edges in the product graph: `pg_graph(q, p, a, n, m)` holds iff the product graph has an
# edge from `(q, n)` to `(p, m)` and the hash value of the edge label is `a`. The program
# also defines two unary relations `pg_source(n, q)` and `pg_target(n, q)` that define the
# source and target nodes of the product graph. The computation of the product graph is
# based on the input finite automaton corresponding to the RPQ and additionally two
# relations `source` and `target` with the sets of source and target nodes of the knowledge
# graph.
#
# The program is used to compute the product graph of the automaton and a knowledge graph.
# The program is used when the user specifies the source and target nodes to be used during
# evaluation.
#
# -----------------------------------------------------------------------------------------
# In the running example, the compiled Rel program follows
#
# // Source nodes of the product graph
# def pg_source(x0, x1): source(x0) x1 = 0
#
# // Target nodes of the product graph
# def pg_target(x0, x1): target(x0) ∧ x1 = 0 ∧ Odd(x0)
# def pg_target(x0, x1): target(x0) ∧ x1 = 1 ∧ Even(x0)
# def pg_target(x0, x1): target(x0) ∧ x1 = 2 ∧ N(x0)
#
# // Edges of the product graph
# def pg_graph(x0, x1, x2, x3, x4): x0 = 0 ∧ x1 = 1 ∧ x2 = hash(-[A]→) ∧ A(x3, x4) ∧ Odd(x3)
# def pg_graph(x0, x1, x2, x3, x4): x0 = 0 ∧ x1 = 2 ∧ x2 = hash(←[B]-) ∧ B(x4, x3)
# def pg_graph(x0, x1, x2, x3, x4): x0 = 1 ∧ x1 = 1 ∧ x2 = hash(-[A]→) ∧ A(x3, x4) ∧ Even(x3)
#
# =========================================================================================


class RPQCompiler(ABC):
    def __init__(self, pattern: RPQ, options):
        self.a = prepare_automaton(pattern, options)

        self.prog_id = next_id()

        self.rules: List[Rule] = []
        self.rel_attrs: Dict[str, List[str]] = dict()

    def yield_node_filters_atoms(self, filters: Iterable[NodeFilter], x: int):
        for f in filters:
            if isinstance(f, NodeLabel):
                yield RelAtom(f.label, (x,))
            elif isinstance(f, AnonymousNodeFilter):
                yield AnonymousFilterAtom(f._unary_function, (x,))

    def yield_edge_transition_atoms(self, t, x, y):
        edge_label = t.get_label()
        if edge_label.direction == 'forward':
            yield RelAtom(edge_label.label, (x, y))
        else:
            yield RelAtom(edge_label.label, (y, x))

        if t.has_anon_edge_filters():
            for f in t.get_anon_edge_filters():
                yield AnonymousFilterAtom(f.binary_function, (x, y))

        if t.has_src_filters():
            yield from self.yield_node_filters_atoms(t.src_filters, x)

        if t.has_trg_filters():
            yield from self.yield_node_filters_atoms(t.trg_filters, y)

    def add_rule(self, head: RelAtom, body: List[Atom], n=None):
        if n is None:
            var_refs = set(head.vars).union([v for a in body for v in a.vars])
            n = len(var_refs)
        self.rules.append(Rule(head, n, body))
        logger.info(f"Added rule: {self.rules[-1]}")

# Prepares (decorates) relations and gathers their definitions.
class ConnCompiler(RPQCompiler):
    def __init__(self, pattern: RPQ, options):
        super().__init__(pattern, options)

        self.edge_transition_id = dict()
        self.final_transition_id = dict()

        self.rules: List[Rule] = []
        self.rel_attrs: Dict[str, List[str]] = dict()

        # self.main_conn_rel is the name of the binary connectivity relation `pq_conn` such
        # that `pq_conn(x, y)` holds iff there is a path from x to y with the automaton
        # having an accepting run (from an initial state to a final state).
        self.main_conn_rel = f'pq_{self.prog_id}_conn'

        # maps state q to the name of a binary connectivity relation `pg_conn_from_q` such
        # that `pg_conn_from_q(x, y)` holds iff there is a path from x to y with the
        # automaton having an accepting run when starting in state q.
        self.conn_rel: Dict[int, str] = dict()

        logger.debug("Initializing PyRel Connectivity Predicate emitter")
        logger.debug(self.a)

        self.prepare_relations()


    # -------------------------------------------------------------------------------------
    # Prepare relations for states, transitions, and final states with necessary decorators
    # -------------------------------------------------------------------------------------
    def prepare_relations(self):
        self.rel_attrs[self.main_conn_rel] = ['@track', '@no_inline']

        for q in self.a.states:
            self.prepare_state_conn_relation(q)

    def prepare_state_conn_relation(self, q):
        if self.a.is_exit(q):
            # the state does not outgoing edge-transitions but only ungrounded final node
            # transitions: we will not create an ungrounded  connectivity relation for it
            # but instead we will always contract it with the preceding transition.
            return

        self.conn_rel[q] = f'pq_{self.prog_id}_conn_from_{q}'
        self.rel_attrs[self.conn_rel[q]] = ['@no_inline']

    # -------------------------------------------------------------------------------------
    # Create rules from the automaton
    # -------------------------------------------------------------------------------------
    def make_rules(self):
        self.make_main_conn_rule()

        for q in self.a.delta:
            for p in self.a.delta[q]:
                for t in self.a.delta[q][p]:
                    self.make_edge_rules(q, p, t)

    def make_main_conn_rule(self):
        q_0 = self.a.get_initial_state()
        if not self.a.is_exit(q_0):
            init_conn_rel = self.conn_rel[q_0]
            self.add_rule(
                head = RelAtom(self.main_conn_rel, (0, 1)),
                body = [RelAtom(init_conn_rel, (0, 1))]
            )
        if q_0 in self.a.final:
            for e in self.a.final[q_0]:
                self.add_rule(
                    head = RelAtom(self.main_conn_rel, (0, 1)),
                    body = [
                        VariableEqualityAtom((0, 1)),
                        *self.yield_node_filters_atoms(e.filters, 0)
                    ]
                )

    def make_edge_rules(self, q, p, t: EdgeTransition):
        if not self.a.is_exit(p):
            self.make_transition_rule(q, p, t)

        if p in self.a.final:
            for e in self.a.final[p]:
                self.make_transition_final_rule(q, t, e)

    def make_transition_rule(self, q, p, t: EdgeTransition):
        assert not self.a.is_exit(p)

        conn_q_rel = self.conn_rel[q]
        conn_p_rel = self.conn_rel[p]

        self.add_rule(
            head = RelAtom(conn_q_rel, (0, 2)),
            body = [
                *self.yield_edge_transition_atoms(t, 0, 1),
                RelAtom(conn_p_rel, (1, 2))
            ]
        )

    def make_transition_final_rule(self, q, t: EdgeTransition, e: NodeTransition):
        conn_q_rel = self.conn_rel[q]

        self.add_rule(
            head = RelAtom(conn_q_rel, (0, 1)),
            body = [
                *self.yield_edge_transition_atoms(t, 0, 1),
                *self.yield_node_filters_atoms(e._filters, 1)
            ]
        )

def compile_conn(pattern, options=DEFAULT_OPTIONS) -> DatalogProgram:
    compiler = ConnCompiler(pattern, options)
    compiler.make_rules()

    program = DatalogProgram(
        root_rel = {'conn_rel': compiler.main_conn_rel},
        rel_attrs = compiler.rel_attrs,
        rules = compiler.rules,
        edge_label_map = None
    )

    return program


class ProductGraphCompiler(RPQCompiler):
    def __init__(self, pattern: RPQ, source_rel: str, target_rel: str, options):
        super().__init__(pattern, options)

        self.source_rel = source_rel
        self.target_rel = target_rel

        # The name of the relation with the edges in the product graph of the knowledge
        # graph and the input automaton. The relation has arity `5` and contains a tuple
        # `(q, p, a, n, m)` iff the product graph has an edge from `(q, n)` to `(p, m)` and
        # the hash value of the edge label is `a`.
        self.pg_graph_rel = f'pg_{self.prog_id}_graph'

        # The relation names of the source and target nodes of the product graph; arity 2,
        # each contains pairs `(n, q)`, where `n` is the node of the knowledge graph and `q`
        # is a state of the automaton.
        self.pg_source_rel = f'pg_{self.prog_id}_source'
        self.pg_target_rel = f'pg_{self.prog_id}_target'

        self.rel_attrs[self.pg_graph_rel] = ['@pipeline']
        self.rel_attrs[self.pg_source_rel] = ['@pipeline']
        self.rel_attrs[self.pg_target_rel] = ['@pipeline']

        # maps hashes of edge labels to edge labels
        self.pg_edge_label_map: Dict[int, EdgeLabel] = {
            hash(lab): lab for lab in self.a.alphabet
        }

        logger.info("Initializing PyRel Product Graph emitter")
        logger.info(self.a)


    # -------------------------------------------------------------------------------------
    # Prepare definitions of the 3 main relations
    # -------------------------------------------------------------------------------------

    def make_rules(self):
        self.make_pg_source_rules()
        self.make_pg_graph_rules()
        self.make_pg_target_rules()

    def make_pg_source_rules(self):
        for q in self.a.init:
            # pg_source(n, x) ← source_rel(n) ∧ x = q
            self.add_rule(
                head = RelAtom(self.pg_source_rel, (0, 1)),
                body = [
                    RelAtom(self.source_rel, (0,)),
                    ConstantEqualityAtom((1, q))
                ]
            )

    def make_pg_target_rules(self):
        for q in self.a.final:
            for e in self.a.final[q]:
                # pg_target(n, x) ← target_rel(n) ∧ x = q ∧ e(n)
                self.add_rule(
                    head = RelAtom(self.pg_target_rel, (0, 1)),
                    body = [
                        RelAtom(self.target_rel, (0,)),
                        ConstantEqualityAtom((1, q)),
                        *self.yield_node_filters_atoms(e.filters, 0)
                    ]
                )

    def make_pg_graph_rules(self):
        for q in self.a.delta:
            for p in self.a.delta[q]:
                for t in self.a.delta[q][p]:
                    # pg_edge(x, y, a, n, m) ← q = x ∧ p = y ∧ a = #label(t) ∧ t(n, m)
                    self.add_rule(
                        head = RelAtom(self.pg_graph_rel, (0, 1, 2, 3, 4)),
                        body = [
                            ConstantEqualityAtom((0, q)),
                            ConstantEqualityAtom((1, p)),
                            ConstantEqualityAtom((2, hash(t.get_label()))),
                            *self.yield_edge_transition_atoms(t, 3, 4)
                        ]
                    )


def compile_product_graph(pattern: RPQ, source_rel: str, target_rel: str, options=DEFAULT_OPTIONS) -> DatalogProgram:
    compiler = ProductGraphCompiler(pattern, source_rel, target_rel, options)
    compiler.make_rules()

    logger.info('Mapping edge labels:')
    for pq_edge_label in compiler.pg_edge_label_map:
            logger.info(f" * {pq_edge_label} is {compiler.pg_edge_label_map[pq_edge_label]}")

    program = DatalogProgram(
        root_rel = {
            'pg_graph_rel': compiler.pg_graph_rel,
            'pg_source_rel': compiler.pg_source_rel,
            'pg_target_rel': compiler.pg_target_rel
        },
        rel_attrs = compiler.rel_attrs,
        rules = compiler.rules,
        edge_label_map = compiler.pg_edge_label_map
    )

    return program



"""
Main functionality for finding paths through automaton.
"""

from relationalai.early_access.builder import define, where, Integer

from .product_graph import create_product_graph #, find_path_in_product_graph
# from .product_graph import project_product_graph_path_to_original_graph

from .path_algorithms.find_paths import find_shortest_paths, find_walks

def find_paths_via_automaton(graph, automaton, Source, Target,
                             num_paths = None, max_length = 1000,
                             type = "shortest"):
    """
    Find a path in the graph allowed by the automaton from Source nodes to Target nodes.

    Args:
        graph: The user's input graph
        automaton: Automaton object representing the pattern to match
        Source: Source node entities
        Target: Target node entities
        num_paths: Optional
          - if specified, limit the number of paths returned
          - if None, all paths are returned.

    Returns:
        product_graph: The product graph
        paths: Relationship of arity 4 encoding paths in the original graph
    """
    model = graph.model
    Node = graph.Node

    # Create product graph:
    product_graph = create_product_graph(graph, automaton)
    ProductGraphNode = product_graph.Node

    # automaton states:
    init_state = 0  # assumes single initial state labeled 0 of the automaton

    FinalState = model.Relationship("final_state1 {Integer}")
    for state in automaton.get_final_states():
        define(FinalState(state))

    # FinalState.inspect()

    # Product graph source and target sets:
    ProductGraphSource = model.Relationship(f"pg_source {{{ProductGraphNode}}}")
    ProductGraphTarget = model.Relationship(f"pg_target {{{ProductGraphNode}}}")

    st = Integer.ref() # state in the automaton
    src, tgt = Node.ref(), Node.ref()

    where(Source(src)).define(
        product_graph_src := ProductGraphNode.new(
            state=init_state, orig_node=src
        ),
        ProductGraphSource(product_graph_src)
    )

    where(Target(tgt), FinalState(st)).define(
        product_graph_dst := ProductGraphNode.new(
            state=st, orig_node=tgt
        ),
        ProductGraphTarget(product_graph_dst)
    )

    # ProductGraphSource.inspect()
    # ProductGraphTarget.inspect()

    if type == "shortest":
        # Find shortest paths in the product graph:
        find_paths_function = find_shortest_paths

    elif type == "walks":
        find_paths_function = find_walks
    else:
        raise ValueError(f"Unknown type of path search: {type}")

    paths = find_paths_function(
        product_graph, ProductGraphSource, ProductGraphTarget,
        num_paths=num_paths, max_length=max_length
    )

    return product_graph, paths



from __future__ import annotations
import atexit
from datetime import datetime, timedelta
import json
import textwrap
import time
from typing import Any, Dict, Tuple, List, cast
from urllib.error import HTTPError

from pandas import DataFrame

from relationalai import debugging
from relationalai.clients.util import poll_with_specified_overhead

from ..errors import EngineNotFoundException, RAIException
from ..rel_utils import assert_no_problems
from ..loaders.loader import emit_delete_import, import_file, list_available_resources
from .config import Config
from .types import EngineState, ImportSource, ImportSourceFile, TransactionAsyncResponse
from .client import Client, ExportParams, ProviderBase, ResourcesBase
from .. import dsl, rel, metamodel as m
from railib import api
from . import result_helpers

#--------------------------------------------------
# Constants
#--------------------------------------------------

TXN_FIELDS = ["id", "account_name", "state", "created_on", "finished_at", "duration", "database_name", "read_only", "engine_name", "query", "query_size", "tags", "user_agent", "response_format_version"]
TXN_REPLACE_MAP = {"database_name": "database", "engine_name": "engine", "account_name": "account", "user_agent": "agent"}
VALID_ENGINE_STATES = ["REQUESTED", "PROVISIONED", "PROVISIONING"]
ENGINE_SIZES = ["XS", "S", "M", "L", "XL"]

#--------------------------------------------------
# Resources
#--------------------------------------------------

class Resources(ResourcesBase):
    def __init__(self, profile:str|None=None, config:Config|None=None):
        super().__init__(profile, config=config)
        self._ctx = None
        atexit.register(self.cancel_pending_transactions)

    def _api_ctx(self):
        if not self._ctx:
            self._ctx = api.Context(**self.config.to_rai_config())
        return self._ctx

    def reset(self):
        self._ctx = None

    #--------------------------------------------------
    # Generic
    #--------------------------------------------------

    def get_version(self):
        raise Exception("Azure version not available")

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    # Note: in contrast to the API definition in `ResourcesBase`, `get_database` in Azure
    # can return a `List` object instead of a `Dict` object.
    def get_database(self, database:str):
        return api.get_database(self._api_ctx(), database)

    # not implemented in Azure
    def get_installed_packages(self, database: str) -> Dict | None:
        return super().get_installed_packages(database)

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------
    def get_engine_sizes(self, cloud_provider: str|None=None):
        return ENGINE_SIZES

    def get_cloud_provider(self) -> str:
        return "azure"

    def list_engines(self, state:str|None = None):
        return api.list_engines(self._api_ctx(), state)

    def get_engine(self, name:str) -> EngineState:
        return cast(EngineState, api.get_engine(self._api_ctx(), name))

    def is_valid_engine_state(self, name:str):
        return name in VALID_ENGINE_STATES

    def create_engine(self, name:str, size:str|None=None, auto_suspend_mins: int|None=None):
        if size is None:
            size = "M"
        with debugging.span("create_engine", name=name, size=size):
            return api.create_engine_wait(self._api_ctx(), name, size)

    def delete_engine(self, name:str, force:bool=False):
        return api.delete_engine(self._api_ctx(), name)

    def suspend_engine(self, name:str):
        return api.suspend_engine(self._api_ctx(), name)

    def resume_engine(self, name:str, headers={}):
        return api.resume_engine_wait(self._api_ctx(), name)

    def resume_engine_async(self, name:str):
        return api.resume_engine(self._api_ctx(), name)

    def auto_create_engine_async(self, name: str | None = None):
        raise Exception("Azure doesn't support auto_create_engine_async")

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    def list_graphs(self) -> List[Any]:
        with debugging.span("list_models"):
            return api.list_databases(self._api_ctx())

    def get_graph(self, name:str):
        with debugging.span("get_model", name=name):
            return api.get_database(self._api_ctx(), name)

    def create_graph(self, name: str):
        with debugging.span("create_model", name=name):
            return api.create_database(self._api_ctx(), name)

    def delete_graph(self, name:str):
        with debugging.span("delete_model", name=name):
            return api.delete_database(self._api_ctx(), name)

    def clone_graph(self, target_name:str, source_name:str, nowait_durable:bool=False):
        # not a typo: the argument order is indeed target then source
        return api.create_database(self._api_ctx(), target_name, source_name)

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        return api.list_databases(self._api_ctx())

    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        rel_code = self.create_models_code(models)
        results = self.exec_raw(database, engine, rel_code, readonly=False)
        if results.problems:
            return results.problems
        return []

    def delete_model(self, database:str, engine:str, name:str):
        return api.delete_model(self._api_ctx(), database, engine, name)

    def create_models_code(self, models:List[Tuple[str, str]]) -> str:
        lines = []
        for (name, code) in models:
            name = name.replace("\"", "\\\"")
            assert "\"\"\"\"\"\"\"" not in code, "Code literals must use fewer than 7 quotes."

            lines.append(textwrap.dedent(f"""
            def delete[:rel, :catalog, :model, "{name}"]: rel[:catalog, :model, "{name}"]
            def insert[:rel, :catalog, :model, "{name}"]: raw\"\"\"\"\"\"\"
            """) + code + "\n\"\"\"\"\"\"\"")
        rel_code = "\n\n".join(lines)
        return rel_code

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        raise Exception("Azure doesn't support exports")

    def create_export(self, params: ExportParams):
        if not params.dry_run:
            raise Exception("Azure doesn't support exports")

    def create_export_table(self, database: str, engine: str, table: str, relation: str, columns: Dict[str, str], code: str, refresh: str|None=None):
        raise Exception("Azure doesn't support exports")

    def delete_export(self, database: str, engine: str, name: str):
        raise Exception("Azure doesn't support exports")

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    def list_imports(self, id:str|None = None, name:str|None = None, model:str|None = None, status:str|None = None, creator:str|None = None):
        if not model:
            raise RAIException("Imports can only be listed for a particular model in azure")
        return [*list_available_resources(self, model, self.get_default_engine_name()).values()]

    def poll_imports(self, sources:List[str], model:str):
        raise Exception("Azure doesn't support import polling")

    def create_import_stream(self, source:ImportSource, model:str, rate = 1, options: dict|None = None):
        raise Exception("Azure doesn't support import streams")

    def create_import_snapshot(self, source:ImportSource, model:str, options: dict|None = None):
        assert isinstance(source, ImportSourceFile), "Azure integration only supports loading from files and URLs right now."
        import_file(self, model, source, **(options or {}))

    def delete_import(self, import_name: str, model:str, force = False):
        res = self.exec_raw(model, self.get_default_engine_name(), emit_delete_import(import_name), False)
        assert_no_problems(res)

    def set_imports_engine_size(self, size: str):
        raise Exception("Azure doesn't support setting imports engine size")

    def change_imports_status(self, suspend:bool):
        raise Exception("Azure doesn't support import status changes")

    def get_imports_status(self):
        return None

    def change_stream_status(self, stream_id: str, model:str, suspend: bool):
        raise Exception("Azure doesn't support stream status changes")

    def get_import_stream(self, name: str|None, model:str|None):
        raise Exception("Azure doesn't support get import streams")

    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    def _exec(self, code:str, params:List[Any]|Any|None = None, raw=False, help=True):
        raise Exception("Azure doesn't support _exec")

    def exec_lqp(self, database: str, engine: str | None, raw_code: bytes, readonly=True, *, inputs: Dict | None = None, nowait_durable=False, headers: Dict | None = None, bypass_index=False):
        raise Exception("Azure doesn't support exec_lqp")

    def exec_raw(self, database:str, engine:str|None, raw_code:str, readonly=True, *, inputs: Dict | None = None, nowait_durable=False, headers: Dict | None = None, raw_results=True):
        if engine is None:
            engine = self.get_default_engine_name()
        try:
            with debugging.span("transaction") as txn_span:
                ctx = self._api_ctx()
                if inputs is None:
                    inputs = {}
                with debugging.span("create"):
                    txn = api.exec_async(ctx, database, engine, raw_code, readonly=readonly, inputs=inputs)
                txn_id = txn.transaction["id"]
                txn_span["txn_id"] = txn_id
                debugging.event("transaction_created", txn_span, txn_id=txn_id)

                # TODO: dedup with SDK
                rsp = api.TransactionAsyncResponse()
                txn = api.get_transaction(ctx, txn_id)
                start_time = time.time()

                def check_done():
                    with debugging.span("check_status"):
                        state = api.get_transaction(ctx, txn_id)["state"]
                        return api.is_txn_term_state(state)

                with debugging.span("wait", txn_id=txn_id):
                    poll_with_specified_overhead(
                        check_done,
                        overhead_rate=0.1,
                        start_time=start_time,
                    )

                # TODO: parallelize
                with debugging.span("fetch"):
                    rsp.transaction = api.get_transaction(ctx, txn_id)
                    rsp.metadata = api.get_transaction_metadata(ctx, txn_id)
                    rsp.problems = api.get_transaction_problems(ctx, txn_id)
                    with debugging.span("fetch_results"):
                        rsp.results = api.get_transaction_results(ctx, txn_id)

                return cast(TransactionAsyncResponse, rsp)
        except HTTPError as err:
            res = json.loads(err.read().decode())
            # Grab request id; useful for searching logs
            request_id = err.headers.get("x-request-id")
            # RAI API uses a JSON payload in the body to explain why the request failed
            # This annotates the error with that to make the exception actually useful.
            if "engine not found" in res.get('message', ''):
                print("") # the SDK appears to print some stuff before the error message
                exception = EngineNotFoundException(cast(str, self.config.get('engine', "Unknown")), res.get('message'))
                raise exception from None
            raise RAIException("HTTPError", res.get('message', ''), f"details: {res.get('details', '')}; request_id: {request_id}")

    def format_results(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        return result_helpers.format_results(results, task)

    #--------------------------------------------------
    # Exec format
    #--------------------------------------------------

    def exec_format(self, database: str, engine: str, raw_code: str, task:m.Task, format:str, inputs: Dict | None = None, readonly: bool = True, nowait_durable=False, headers: Dict | None = None) -> Any: # @FIXME: Better type annotation
        raise Exception("Azure doesn't support alternative formats yet")

    def to_model_type(self, model:dsl.Graph, name: str, source:str):
        raise Exception("Azure doesn't support import types yet")

    #--------------------------------------------------
    # Transactions
    #--------------------------------------------------

    def get_transaction(self, transaction_id):
        txn = api.get_transaction(self._api_ctx(), transaction_id)
        if not txn:
            return None
        created_on = txn.get("created_on")
        finished_at = txn.get("finished_at")
        duration = txn.get("duration")
        if duration:
            txn["duration"] = timedelta(milliseconds=duration)
        elif created_on:
            txn["duration"] = datetime.now() - datetime.fromtimestamp(created_on / 1000)
        if created_on:
            txn["created_on"] = datetime.fromtimestamp(created_on / 1000)
        if finished_at:
            txn["finished_at"] = datetime.fromtimestamp(finished_at / 1000)
        # Remap based on the fields we care about
        result = {TXN_REPLACE_MAP.get(k, k): v for k, v in txn.items() if k in TXN_FIELDS}
        return result

    def remap_fields(self, transactions):
        if not transactions:
            return []
        for transaction in transactions:
            for key in list(transaction.keys()):
                if key in TXN_REPLACE_MAP:
                    transaction[TXN_REPLACE_MAP[key]] = transaction.pop(key)
        return transactions

    def list_transactions(self, **kwargs):
        TERMINAL_STATES = ["COMPLETED", "ABORTED"]
        VALID_KEYS = ["id", "state", "engine"]

        state = kwargs.get("state")
        only_active = kwargs.get("only_active", False)
        options = {}

        # Azure sdk supports more than just VALID_KEYS as filters but for now we pass through those
        for k, v in kwargs.items():
            if k in VALID_KEYS and v is not None:
                # Only pass state if it is a valid terminal state
                if k == "state" and v.upper() in TERMINAL_STATES:
                    options[k] = v.upper()
                if k != "state":
                    if k == "engine":
                        k = "engine_name"
                    options[k] = v
        # In Azure we store transactions in cosmos and consul
        # Cosmos if the state is terminal (COMPLETED or ABORTED) and Consul if the state is not (e.g. "RUNNING")
        # So we can not filter on active non terminal states via the options passed
        transactions = api.list_transactions(self._api_ctx(), **options)

        if not transactions:
            return []
        # We filter non terminal transactions here
        if only_active:
            transactions = [t for t in transactions if t["state"] in ["CREATED", "RUNNING", "PENDING"]]
        if (isinstance(state, str) and state.upper() not in TERMINAL_STATES):
            transactions = [t for t in transactions if t["state"] in [state.upper()]]
        return self.remap_fields(transactions)

    def cancel_transaction(self, transaction_id):
        return api.cancel_transaction(self._api_ctx(), transaction_id)

    def cancel_pending_transactions(self):
        # all transactions are executed synchronously against azure
        pass

    def get_transaction_events(self, transaction_id:str, continuation_token:str):
        return api._get_resource(
            self._api_ctx(),
            f"/transactions/{transaction_id}/events/profiler?continuation_token={continuation_token}",
        )

    def is_account_flag_set(self, flag: str):
        raise Exception("Azure doesn't support account flags")

    def is_direct_access_enabled(self) -> bool:
      raise Exception("Azure doesn't support direct access")
#--------------------------------------------------
# Provider
#--------------------------------------------------

class Provider(ProviderBase):

    def __init__(
        self,
        profile: str | None = None,
        config: Config | None = None,
        resources: Resources | None = None,
    ):
        if resources:
            self.resources = resources
        else:
            self.resources = Resources(profile, config)

#--------------------------------------------------
# Graph
#--------------------------------------------------

def Graph(name, *, profile:str|None=None, config:Config, dry_run:bool=False, isolated=True, keep_model:bool=False, format="default"):
    use_monotype_operators = config.get("compiler.use_monotype_operators", False)

    client = Client(
        Resources(profile, config),
        rel.Compiler(config),
        name,
        config,
        dry_run=dry_run,
        isolated=isolated,
        keep_model=keep_model,
    )
    base_rel = """
        @inline
        def make_identity(x..., z):
            rel_primitive_hash_tuple_uint128(x..., z)

        @inline
        def pyrel_default({F}, c, k..., v):
            F(k..., v) or (not F(k..., _) and v = c)

        @inline
        def pyrel_unwrap(x in UInt128, y): y = x

        @inline
        def pyrel_dates_period_days(x in Date, y in Date, z in Int):
            exists((u) | dates_period_days(x, y , u) and u = ^Day[z])

        @inline
        def pyrel_datetimes_period_milliseconds(x in DateTime, y in DateTime, z in Int):
            exists((u) | datetimes_period_milliseconds(x, y , u) and u = ^Millisecond[z])

        @inline
        def pyrel_bool_filter(a, b, {F}, z): { z = if_then_else[F(a, b), boolean_true, boolean_false] }

        @inline
        def pyrel_strftime(v, fmt, tz in String, s in String):
            (Date(v) and s = format_date[v, fmt])
            or (DateTime(v) and s = format_datetime[v, fmt, tz])

        @inline
        def pyrel_regex_match_all(pattern, string in String, pos in Int, offset in Int, match in String):
            regex_match_all(pattern, string, offset, match) and offset >= pos

        @inline
        def pyrel_regex_match(pattern, string in String, pos in Int, offset in Int, match in String):
            pyrel_regex_match_all(pattern, string, pos, offset, match) and offset = pos

        @inline
        def pyrel_regex_search(pattern, string in String, pos in Int, offset in Int, match in String):
            enumerate(pyrel_regex_match_all[pattern, string, pos], 1, offset, match)

        @inline
        def pyrel_regex_sub(pattern, repl in String, string in String, result in String):
            string_replace_multiple(string, {(last[regex_match_all[pattern, string]], repl)}, result)

        @inline
        def pyrel_capture_group(regex in Pattern, string in String, pos in Int, index, match in String):
            (Integer(index) and capture_group_by_index(regex, string, pos, index, match)) or
            (String(index) and capture_group_by_name(regex, string, pos, index, match))

        declare __resource
        declare __compiled_patterns
    """
    if use_monotype_operators:
        base_rel += """

        // use monotyped operators
        from ::std::monotype import +, -, *, /, <, <=, >, >=
        """
    pyrel_base = dsl.build.raw_task(base_rel)
    debugging.set_source(pyrel_base)
    client.install("pyrel_base", pyrel_base)
    return dsl.Graph(client, name, format=format)



from relationalai.early_access.dsl.core.constraints import Constraint
from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.logic import LogicFragment
from relationalai.early_access.dsl.core.logic.helper import one_point_reduce

class ExistentialConstraint(Constraint, LogicFragment):

    # Each Schema object is a ContextManager
    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()

    def __init__(self):
        LogicFragment.__init__(self)

    def grounds(self, v, groundings):

        # First, we need to extend grounds to include variables introduce in this constraint
        prop_vars = [d for d in self._scalars.values() if not d.relational()]

        addl = {}
        glen = -1

        props = self.propositional_constraints()

        while glen < len(addl):
            glen = len(addl)

            for c in props.values():
                for pv in prop_vars:
                    if c.grounds(pv, dict(groundings, **addl)):
                        addl[pv.display()] = pv

        gdgs = dict(groundings, **addl)

        for c in props.values():
            if c.grounds(v, gdgs):
                return True

        return False

    def free_variables(self):
        fvars = {}
        bvars = self._scalars
        for c in self.propositional_constraints().values():
            cdict = c.scalar_refs()
            for v in cdict:
                if v not in bvars:
                    fvars[v] = cdict[v]
        return fvars

    # Returns a new ExistentialConstraint that duplicates self after
    # mapping every Relation referred to by an atom through relmap
    #
    def map(self, relmap):
        x = ExistentialConstraint()
        x._scalars = self._scalars
        x._sconstraints = self._sconstraints
        for a in self._atoms.values():
            newatom = a.map(relmap)
            x._atoms[newatom.entityid()] = newatom
        for ex in self._existentials.values():
            newex = ex.map(relmap)
            x._existentials[newex.entityid()] = newex
        return x

    # Returns a new ExistentialConstraint that duplicates self after
    # replacing each atom with the one constructed by the given builder
    # function taking the relational component and arguments of the current atom.
    #
    def map_builder(self, builder):
        x = ExistentialConstraint()
        x._scalars = self._scalars      
        x._sconstraints = {}
        for a in self._sconstraints.values():
            ap = a.map_builder(builder)
            x._sconstraints[ap.entityid()] = ap

        for a in self._atoms.values():
            new_atom = a.map_builder(builder)
            x._atoms[new_atom.entityid()] = new_atom
        for ex in self._existentials.values():
            new_ex = ex.map_builder(builder)
            x._existentials[new_ex.entityid()] = new_ex
        return x

    @staticmethod
    def build_existential() -> 'ExistentialConstraint':
        return ExistentialConstraint()

    @staticmethod
    def assemble_existential(decls, comps, atoms) -> 'ExistentialConstraint':
        x = ExistentialConstraint.build_existential()
        x._scalars = decls
        x._sconstraints = comps
        x._atoms = atoms
        return x

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        vdecl = "; ".join([v for v in self._scalars])
        head = f"exists(({vdecl}) |"

        props = [f"    {c.pprint()}" for c in self.propositional_constraints().values()]
        body = " true )" if len(props) == 0 else "\n" + "\n".join(props) + " )"

        return head + body

    def scalar_refs(self):
        return self.free_variables()

    def refersto(self, varname: str) -> bool:
        return varname in self.free_variables()

    def rel_formula(self):
        propcs = self.propositional_constraints()

        if len(self._scalars) + len(propcs) == 0:
            return "true"

        props = [f"    {c.rel_formula()}" for c in propcs.values()]

        if len(self._scalars) > 0:
            vdecl = ", ".join([v for v in self._scalars])
            head = f"exists(({vdecl}) |"
            if len(props) == 0:
                return head + " true)"
            else:
                return head + "\n" + " and\n".join(props) + " )"
        else:
            if len(props) == 0:
                return ""
            else:
                return "(" + " and\n".join(props) + " )"

    def rename(self, renaming):

        new_vars = {}
        for d in self._scalars:
            dv = self._scalars[d]
            if d in renaming:
                new_vars[d] = dv.rename(renaming[d])
            else:
                new_vars[d] = dv

        new_comps = {}
        for a in self._sconstraints.values():
            ap = a.rename(renaming)
            new_comps[ap.entityid()] = ap

        new_atoms = {}
        for a in self._atoms.values():
            ap = a.rename(renaming)
            new_atoms[ap.entityid()] = ap

        return self.assemble_existential(new_vars, new_comps, new_atoms)

    def simplify(self):
        comps = {}
        for p in self._sconstraints.values():
            ps = p.simplify()
            comps[ps.entityid()] = ps

        atoms = {}
        for p in self._atoms.values():
            ps = p.simplify()
            atoms[ps.entityid()] = ps

        (decls, c, a) = one_point_reduce(self._scalars, comps, atoms)
        return self.assemble_existential(decls, c, a)

    def substitute(self, bindings):
        decls = {}
        for d in self._scalars.values():
            dp = d.substitute(bindings)
            decls[dp.display()] = dp

        comps = {}
        for p in self._sconstraints.values():
            pp = p.substitute(bindings)
            comps[pp.entityid()] = pp

        atoms = {}
        for p in self._atoms.values():
            pp = p.substitute(bindings)
            atoms[pp.entityid()] = pp

        return self.assemble_existential(decls, comps, atoms)

    def entityid(self):
        return hash((hash(tuple(self._scalars)),
                     hash(tuple(self._sconstraints)),
                     hash(tuple(self._atoms))))

# Given a LogicFragment of the form:
#
#     (bvars: sconsts & atoms)
#
# where:
#
#   - bvars is a set of ScalarVariable declarations,
#   - sconsts is a dictionary of ScalarConstraints, and
#   - atoms is a dictionary of AtomicConstraints
#
# Existentially quantify out all of the variables in bvars, returning either
# a modified pair (sconsts', atoms') or an ExistentialConstraint that introduces
# any remaining variables bvarsPrime under a quantifier.
#
# Note that constraints in sconsts and/or atoms could refer to ScalarVariables
# that are not in bvars. In this case, those variables will occur free in the
# fragment returned.
#
def existentially_quantify(bvars, sconsts, atoms):
    (n, c, a) = one_point_reduce(bvars, sconsts, atoms)

    if len(n) > 0:
        return {}, {}, ExistentialConstraint.assemble_existential(n, c, a)
    else:
        return c, a, None


from relationalai.early_access.builder import Integer
from relationalai.early_access.builder import define, not_

from ..graph import Graph
from ..utilities.iterators import setup_iteration


# Source and Target are concepts representing sets of nodes
def ball_upto(g:Graph, Source, Target, max_length=None):
    edge = g.Edge
    Node = g.Node

    # ball(d, ⋅) is the sphere of nodes at distance d from the source node
    # visited(d, ⋅) is the union of all balls up to distance d

    ball = g.model.Relationship(f"ball1 {{Integer}} {{{Node}}}") # , [Integer, Node])
    visited = g.model.Relationship(f"visited1 {{{Node}}}") # , [Integer, Node])
    condition = g.model.Relationship("condition1 {Integer}")

    if max_length is None:
        iter = setup_iteration(g.model, condition, 0, 1000000)
    else:
        iter = setup_iteration(g.model, condition, 0, max_length)

    src, tgt = Node.ref(), Node.ref()

    # Ball around src contains src at distance 0:
    define(ball(0, src)).where(Source(src))

    u, v = g.Node.ref(), g.Node.ref()
    n, k = Integer.ref(), Integer.ref()

    # Recursive case:
    define(ball(k, v)).where(
        iter(k),
        ball(k - 1, u),
        edge(u, v),
        not_(
            visited(v),
        ),
    )

    define(ball(k, u)).where(
        ball(k, u)
    )

    define(visited(u)).where(
        iter(k),
        ball(k, u)
    )

    define(visited(u)).where(
        visited(u)
    )

    define(condition(n)).where(
        iter(n),
        ball(n, u),
        not_(
            ball(n, tgt),
            Target(tgt)
        )
    )

    return ball


def ball_upto_alt(g:Graph, Source, Target, max_length=None):
    edge = g.Edge
    Node = g.Node

    # ball(d, ⋅) is the sphere of nodes at distance d from the source node

    ball = g.model.Relationship(f"ball2 {{Integer}} {{{Node}}}") # , [Integer, Node])
    condition = g.model.Relationship("condition2 {Integer}")

    if max_length is None:
        iter = setup_iteration(g.model, condition, 0, 1000000)
    else:
        iter = setup_iteration(g.model, condition, 0, max_length)

    src, tgt = Node.ref(), Node.ref()

    # Ball around src contains src at distance 0:
    define(ball(0, src)).where(Source(src))

    u, v = g.Node.ref(), g.Node.ref()
    n, k = Integer.ref(), Integer.ref()

    # Recursive case:
    define(ball(k, v)).where(
        iter(k),
        ball(k - 1, u),
        edge(u, v),
        not_(
            ball(n, v),
        ),
    )

    define(ball(k, u)).where(
        ball(k, u)
    )

    define(condition(n)).where(
        iter(n),
        ball(n, u),
        not_(
            ball(n, tgt),
            Target(tgt)
        )
    )

    return ball
